/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static de.ozgcloud.vorgang.command.PersistedCommand.*;
import static de.ozgcloud.vorgang.common.db.MongoDbUtil.*;
import static de.ozgcloud.vorgang.common.db.CriteriaUtil.*;
import static org.springframework.data.mongodb.core.query.Criteria.*;
import static org.springframework.data.mongodb.core.query.Query.*;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.FindAndModifyOptions;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.data.mongodb.core.query.UpdateDefinition;
import org.springframework.stereotype.Repository;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;

@Repository
class CommandRepository {

	private static final String MONGODB_ASSIGNED_TO = "body.assignedTo";
	private static final String MONGODB_PARENT_ID = BODY_OBJECT_FIELD + "." + PersistedCommand.PROPERTY_PARENT_ID;
	private static final String MONGODB_COMPLETE_IF_SUBS_COMPLETED = BODY_OBJECT_FIELD + "."
			+ PersistedCommand.PROPERTY_COMPLETE_IF_SUBS_COMPLETED;
	private static final String MONGODB_CLASS = "_class";

	@Autowired
	private MongoOperations mongoOperations;

	public Command save(Command command) {
		return mongoOperations.save(command);
	}

	void finishCommand(String commandId) {
		finishCommand(commandId, new Update());
	}

	void finishCommand(String commandId, String createdResource) {
		finishCommand(commandId, new Update().set(PersistedCommand.CREATED_RESOURCE_FIELD, createdResource));
	}

	private void finishCommand(String commandId, Update update) {
		var enhancedUpdate = update.set(MONGODB_STATUS, CommandStatus.FINISHED).set(FINISHED_AT_FIELD, ZonedDateTime.now());
		updateFirst(queryByIdAndPending(commandId), enhancedUpdate);
	}

	private Query queryByIdAndPending(String commandId) {
		return query(new Criteria().andOperator(
				whereIdIs(commandId),
				isStatusIn(CommandStatus.PENDING, CommandStatus.REVOKE_PENDING)));
	}

	void updateCommandStatus(String commandId, CommandStatus status) {
		updateFirstById(commandId, new Update().set(MONGODB_STATUS, status));
	}

	public Command updateCommandStatusAndReturnPrevious(String commandId, CommandStatus status) {
		return mongoOperations.findAndModify(queryById(commandId), new Update().set(MONGODB_STATUS, status), Command.class);
	}

	void setErrorMessage(String commandId, String errorMsg) {
		var update = new Update()
				.set(MONGODB_STATUS, CommandStatus.ERROR)
				.set(ERROR_MSG_FIELD, errorMsg)
				.set(FINISHED_AT_FIELD, ZonedDateTime.now());
		updateFirstById(commandId, update);
	}

	void updateCommandStatusAndVersion(String commandId, CommandStatus status, long relationVersion) {
		var update = new Update()
				.set(MONGODB_STATUS, status)
				.set(RELATION_VERSION_FIELD, relationVersion);
		updateFirstById(commandId, update);
	}

	Optional<Command> findById(String commandId) {
		return Optional.ofNullable(mongoOperations.findById(commandId, Command.class));
	}

	boolean existsPendingCommands(String vorgangId) {
		return exists(buildPendingCommandQuery(vorgangId));
	}

	List<Command> getPendingCommands(String relationId) {
		return find(buildPendingCommandQuery(relationId));
	}

	private Query buildPendingCommandQuery(String vorgangId) {
		return query(new Criteria().andOperator(isVorgangId(vorgangId),
				isStatusIn(CommandStatus.PENDING.name(), CommandStatus.REVOKE_PENDING.name())));
	}

	public List<Command> findCommands(String vorgangId, Collection<CommandStatus> status, Optional<Order> order) {
		return find(buildCommandsQuery(vorgangId, status, order));
	}

	private List<Command> find(Query query) {
		return mongoOperations.find(query, Command.class);
	}

	private Query buildCommandsQuery(String vorgangId, Collection<CommandStatus> status, Optional<Order> commandOrder) {
		var result = query(isVorgangId(vorgangId));

		if (!status.isEmpty()) {
			result.addCriteria(isStatusIn(status));
		}
		commandOrder.map(Order::name).ifPresent(order -> result.addCriteria(isOrder(order)));

		return result;
	}

	public Stream<Command> findCommands(Order order, CommandStatus status, ZonedDateTime createdBefore) {
		return mongoOperations.stream(buildCommandsQuery(order, status, createdBefore), Command.class);
	}

	private Query buildCommandsQuery(Order order, CommandStatus status, ZonedDateTime createdBefore) {
		var result = query(where(ORDER_FIELD).is(order.name()).and(MONGODB_STATUS).is(status));
		result.addCriteria(new Criteria(CREATED_AT_FIELD).lte(createdBefore));
		return result;
	}

	public void patch(String commandId, Map<String, Object> previousState) {
		updateFirstById(commandId, buildUpdateFieldsByMap(PersistedCommand.PREVIOUS_STATE_FIELD, previousState));
	}

	public void deleteAllByVorgang(String vorgangId) {
		mongoOperations.remove(query(where(VORGANG_ID_FIELD).is(vorgangId)), Command.COLLECTION_NAME);
	}

	public boolean existsCommandWithUserId(String userId) {
		return exists(query(buildCreatedByOrAssignedToCriteria(userId)));
	}

	Criteria buildCreatedByOrAssignedToCriteria(String userId) {
		return new Criteria().orOperator(where(CREATED_BY_FIELD).is(userId), where(MONGODB_ASSIGNED_TO).is(userId));
	}

	public void addSubCommands(String commandId, Map<String, Object> subCommandValues) {
		updateFirstById(commandId, buildUpdateFieldsByMap(BODY_OBJECT_FIELD, subCommandValues));
	}

	private Update buildUpdateFieldsByMap(String field, Map<String, Object> values) {
		var update = new Update();
		values.forEach((key, value) -> update.set("%s.%s".formatted(field, key), value));
		return update;
	}

	private void updateFirstById(String commandId, Update update) {
		updateFirst(queryById(commandId), update);
	}

	private void updateFirst(Query query, Update update) {
		mongoOperations.updateFirst(query, update, Command.class);
	}

	public boolean isCommandFailed(String commandId) {
		return exists(buildCommandIsErrorQuery(commandId));
	}

	private Query buildCommandIsErrorQuery(String commandId) {
		return query(new Criteria().andOperator(whereIdIs(commandId), where(MONGODB_STATUS).is(CommandStatus.ERROR)));
	}

	public Optional<String> getParentId(String commandId) {
		var query = queryById(commandId);
		query.fields().include(MONGODB_CLASS, MONGODB_PARENT_ID);
		return findOne(query).map(Command::getBodyObject).map(this::getParentIdProperty);
	}

	private Criteria whereIdIs(String commandId) {
		return where(ID_FIELD).is(commandId);
	}

	String getParentIdProperty(Map<String, Object> body) {
		return MapUtils.getString(body, PersistedCommand.PROPERTY_PARENT_ID);
	}

	Optional<Command> findOne(Query query) {
		return Optional.ofNullable(mongoOperations.findOne(query, Command.class));
	}

	public boolean isCompleteIfSubsCompleted(String commandId) {
		return exists(buildQueryCompleteIfSubsCompleted(commandId));
	}

	Query buildQueryCompleteIfSubsCompleted(String commandId) {
		return query(new Criteria().andOperator(isId(commandId), where(MONGODB_COMPLETE_IF_SUBS_COMPLETED).is(true)));
	}

	public boolean existsNotFinishedSubCommands(String parentId) {
		return exists(buildQueryNotInStatus(parentId, CommandStatus.FINISHED, CommandStatus.REVOKED));
	}

	private boolean exists(Query query) {
		return mongoOperations.exists(query, Command.class);
	}

	Query buildQueryNotInStatus(String parentId, CommandStatus status, CommandStatus... statuses) {
		var excludeStatuses = ArrayUtils.<Object>add(statuses, status);
		return query(new Criteria().andOperator(
				buildIsParentIdCriteria(parentId),
				where(MONGODB_STATUS).nin(excludeStatuses)));
	}

	public Stream<String> findSubCommandIds(String parentId) {
		return mongoOperations.stream(includeIdAndClass(query(buildIsParentIdCriteria(parentId))), Command.class).map(Command::getId);
	}

	private Criteria buildIsParentIdCriteria(String parentId) {
		return where(MONGODB_PARENT_ID).is(parentId);
	}

	Query includeIdAndClass(Query query) {
		query.fields().include(MONGODB_CLASS, ID_FIELD);
		return query;
	}

	public Command setRevokeStatus(String id) {
		return findAndModifyById(id, buildUpdateStatusRevoke());
	}

	private UpdateDefinition buildUpdateStatusRevoke() {
		return buildConditionalStatusUpdateDefinition(
				createStatusCaseOperator(CommandStatus.NEW, CommandStatus.CANCELED),
				createStatusCaseOperator(CommandStatus.PENDING, CommandStatus.REVOKE_PENDING),
				createStatusCaseOperator(CommandStatus.FINISHED, CommandStatus.REVOKE_PENDING));
	}

	public Optional<Command> setRevokeStatusIfNotPending(String id) {
		return Optional.ofNullable(findAndModifyById(id, buildUpdateStatusRevokeIfNotPending()));
	}

	private Command findAndModifyById(String id, UpdateDefinition update) {
		return findAndModify(queryById(id), update);
	}

	private Command findAndModify(Query query, UpdateDefinition update) {
		return mongoOperations.findAndModify(query, update, FindAndModifyOptions.options().returnNew(true), Command.class);
	}

	private Query queryById(String commandId) {
		return query(whereIdIs(commandId));
	}

	private UpdateDefinition buildUpdateStatusRevokeIfNotPending() {
		return buildConditionalStatusUpdateDefinition(
				createStatusCaseOperator(CommandStatus.NEW, CommandStatus.CANCELED),
				createStatusCaseOperator(CommandStatus.FINISHED, CommandStatus.REVOKE_PENDING));
	}

}