/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.db;

import static de.ozgcloud.vorgang.common.db.MongoDbUtil.*;
import static de.ozgcloud.vorgang.vorgang.Vorgang.*;
import static org.springframework.data.mongodb.core.query.Criteria.*;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.springframework.data.mongodb.core.query.Criteria;

import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;

@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class CriteriaUtil {

	private static final String CASE_INSENSITIV = "i";

	public static final String DEFAULT_CLIENT = "Alfa";
	public static final String CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST = "nextWiedervorlageFrist";
	public static final String CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_KEY = String.format("%s.%s.%s", Vorgang.FIELD_CLIENT_ATTRIBUTES,
			DEFAULT_CLIENT, CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST);
	public static final String CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_VALUE_KEY = CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_KEY + ".value";

	public static Criteria isId(String id) {
		return where(MONGODB_FIELDNAME_ID).is(id);
	}

	public static Criteria isNotDeleted() {
		return where(MONGODB_FIELDNAME_STATUS).ne(Status.DELETED);
	}

	public static Criteria inOrganisationseinheitenIds(Collection<String> ids) {
		return where(FIELD_ORGANISATIONSEINHEIT).in(ids);
	}

	public static Criteria vorgangNotInCreation() {
		return where(MONGODB_FIELDNAME_IN_CREATION).is(false);
	}

	public static Criteria vorgangIsAssignedTo(String assignedTo) {
		return new Criteria(MONGODB_FIELDNAME_ASSIGNED_TO).is(assignedTo);
	}

	public static Criteria vorgangIsUnassigned() {
		return where(MONGODB_FIELDNAME_ASSIGNED_TO).exists(false);
	}

	public static Criteria vorgangInStatus(List<Status> status) {
		return where(MONGODB_FIELDNAME_STATUS).in(status);
	}

	public static Criteria vorgangRemainingStatus() {
		return where(MONGODB_FIELDNAME_STATUS).not().in(Status.NEU, Status.ANGENOMMEN, Status.BESCHIEDEN, Status.IN_BEARBEITUNG,
				Status.WEITERGELEITET);
	}

	public static Criteria vorgangStatusBearbeitet() {
		return where(MONGODB_FIELDNAME_STATUS).in(Status.BESCHIEDEN, Status.WEITERGELEITET);
	}

	public static Criteria vorgangStatusZuBearbeitendeEA() {
		return where(MONGODB_FIELDNAME_STATUS).in(Status.NEU, Status.ANGENOMMEN);
	}

	public static Criteria vorgangStatusZuBearbeitende() {
		return where(MONGODB_FIELDNAME_STATUS).in(Status.NEU, Status.ANGENOMMEN, Status.IN_BEARBEITUNG);
	}

	public static Criteria vorgangStatusInBearbeitung() {
		return where(MONGODB_FIELDNAME_STATUS).is(Status.IN_BEARBEITUNG);
	}

	public static Criteria searchCriteria(String searchString) {
		var searchParts = searchString.trim().split(StringUtils.SPACE);
		return new Criteria().andOperator(Arrays.stream(searchParts).map(CriteriaUtil::buildSearchPartCriteria).toList());
	}

	private static Criteria buildSearchPartCriteria(String searchPart) {
		return new Criteria().orOperator(
				new Criteria(FIELD_EINGANGSKENNZ).regex(searchPart, CASE_INSENSITIV),
				new Criteria(FIELD_AKTENZEICHEN).regex(searchPart, CASE_INSENSITIV),
				new Criteria(FIELD_ANTRAGSTELLER_VORNAME).regex(searchPart, CASE_INSENSITIV),
				new Criteria(FIELD_ANTRAGSTELLER_NACHNAME).regex(searchPart, CASE_INSENSITIV),
				new Criteria(FIELD_NAME).regex(searchPart, CASE_INSENSITIV),
				new Criteria(FIELD_NUMMER).regex(searchPart, CASE_INSENSITIV));
	}

	public static Criteria nextWiedervorlageFristInThePast() {
		return where(CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_VALUE_KEY).lte(LocalDate.now().toString());
	}

	public static Criteria nextWiedervorlageFristInTheFuture() {
		return where(CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_VALUE_KEY).gt(LocalDate.now().toString());
	}

	public static Criteria withoutNextWiedervorlageFrist() {
		return Criteria.where(CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_VALUE_KEY).isNull();
	}

	public static Criteria nextWiedervorlageFristExists() {
		return Criteria.where(CriteriaUtil.CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_KEY).exists(true);
	}

	public static Criteria whereIdAndVersion(String vorgangId, long version) {
		return isId(vorgangId).and(MONGODB_FIELDNAME_VERSION).is(version);
	}

	public static Criteria isVorgangId(String vorgangId) {
		return where(PersistedCommand.VORGANG_ID_FIELD).is(vorgangId);
	}

	public static Criteria isStatusIn(Object... status) {
		return where(MONGODB_STATUS).in(status);
	}

	public static Criteria isOrder(String order) {
		return where(PersistedCommand.ORDER_FIELD).is(order);
	}
}