/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.client.match.MockRestRequestMatchers.*;
import static org.springframework.test.web.client.response.MockRestResponseCreators.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.web.client.ExpectedCount;
import org.springframework.test.web.client.MockRestServiceServer;
import org.springframework.web.client.RestTemplate;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.nachrichten.attributes.ClientAttributeService;
import de.ozgcloud.nachrichten.postfach.osi.MessageAttachmentTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.MessageJsonReplyTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.OsiPostfachProperties;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.files.FileId;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.files.OzgFile;
import io.grpc.stub.StreamObserver;

@SpringBootTest(classes = { VorgangManagerServerApplication.class, OsiPostfachProperties.class }, properties = {
		"grpc.server.in-process-name=postfachitcase",
		"ozgcloud.osi.postfach.proxyapi.url=http://localhost/ApiProxy/V1/Message",
		"ozgcloud.osi.postfach.proxyapi.key=1234",
		"ozgcloud.osi.postfach.proxyapi.realm=test-realm",
		"ozgcloud.osi.postfach.notification.mail-from=test@local.host",
		"grpc.client.ozgcloud-command-manager.address=in-process:postfachitcase",
		"grpc.client.command-manager.address=in-process:postfachitcase",

})
@WithMockUser
@DataITCase
@DirtiesContext(classMode = DirtiesContext.ClassMode.BEFORE_CLASS)
class PostfachMailITCase {

	private static final String TEST_OSI_POSTFACH_URI = "http://localhost/ApiProxy/V1/Message";

	@Autowired
	private PostfachGrpcService grpcService;

	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private RestTemplate restTemplate;
	@MockBean
	private PolicyService policyService;

	@Mock
	private StreamObserver<GrpcFindPostfachMailsResponse> responseObserver;
	@Captor
	private ArgumentCaptor<GrpcFindPostfachMailsResponse> responseCaptor;

	@MockBean
	private ClientAttributeService clientAttributeService;

	@SpyBean
	private CommandService commandService;
	private MockRestServiceServer mockServer;

	@BeforeEach
	void prepareDatabase() {
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);

		when(policyService.hasOrganisationEinheitenId(any())).thenReturn(true);
		when(policyService.hasOrganisationEinheitenIdByVorgangAttachedItem(any())).thenReturn(true);
	}

	@BeforeEach
	void initRestMock() {
		mockServer = MockRestServiceServer.createServer(restTemplate);
	}

	@Nested
	class TestSendingMail {

		@Mock
		private StreamObserver<GrpcSendPostfachMailResponse> responseObserver;
		@Mock
		private StreamObserver<GrpcResendPostfachMailResponse> resendResponseObserver;

		@Test
		void shouldCallOsiPostfachServer() {
			mockServerSendSuccess();

			callGrpcEndpoint();

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> mockServer.verify());
		}

		@Nested
		class TestSendAndListMail {

			@Test
			void shouldListSentMails() {
				mockServerSendSuccess();

				callGrpcEndpoint();

				var mails = await().atMost(60, TimeUnit.SECONDS)
						.until(PostfachMailITCase.this::callGrpcListEndpoint, PostfachMailITCase.this::hasAtLeastOneElement);

				assertThat(mails.get(0).getCreatedAt()).isNotNull();
				assertThat(mails.get(0).getDirection()).isEqualTo(GrpcDirection.OUT);
				assertThat(mails.get(0).getSentAt()).isNotNull();
				assertThat(ZonedDateTime.parse(mails.get(0).getSentAt())).isCloseTo(ZonedDateTime.now(), within(61, ChronoUnit.SECONDS));
				assertThat(mails.get(0).getSentSuccessful()).isTrue();
			}
		}

		@Nested
		class TestSendingFailed {

			@BeforeEach
			void init() {
				mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
			}

			@Test
			void shouldMarkFailed() {
				mockServerSendFailed();
				assertThat(callGrpcListEndpoint()).isEmpty();

				callGrpcEndpoint();

				await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
					assertThat(callGrpcListEndpoint()).hasSize(1).first().satisfies(mail -> {
						assertThat(mail.getSentAt()).isNotNull();
						assertThat(ZonedDateTime.parse(mail.getSentAt())).isCloseTo(ZonedDateTime.now(), within(61, ChronoUnit.SECONDS));
						assertThat(mail.getSentSuccessful()).isFalse();
					});
				});
			}

			@Test
			void resendWithSuccess() {
				mockServerSendFailed();
				mockServerSendSuccess();
				callGrpcEndpoint();

				var mails = await().atMost(60, TimeUnit.SECONDS)
						.until(PostfachMailITCase.this::callGrpcListEndpoint, PostfachMailITCase.this::hasAtLeastOneElement);
				var mailId = mails.get(0).getId();

				callResendGrpcEndpoint(mailId);
				await().atMost(60, TimeUnit.SECONDS).untilAsserted(
						() -> assertThat(callGrpcListEndpoint()).hasSize(1).first().extracting(GrpcPostfachMail::getSentSuccessful).isEqualTo(true));
			}

			@Test
			void resentFailed() {
				mockServerSendFailed();
				mockServerSendFailed();

				callGrpcEndpoint();
				var mails = await().atMost(60, TimeUnit.SECONDS)
						.until(PostfachMailITCase.this::callGrpcListEndpoint, PostfachMailITCase.this::hasAtLeastOneElement);
				var mailId = mails.get(0).getId();

				callResendGrpcEndpoint(mailId);

				mails = callGrpcListEndpoint();
				assertThat(mails.get(0).getSentSuccessful()).isFalse();
				assertThat(mails.get(0).getMessageCode()).isEqualTo(PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE.getMessageCode());
			}
		}

		private void mockServerSendFailed() {
			mockServer.expect(ExpectedCount.once(), requestTo(TEST_OSI_POSTFACH_URI))
					.andExpect(method(HttpMethod.POST))
					.andRespond(withSuccess("false", MediaType.APPLICATION_JSON));
		}

		private void mockServerSendSuccess() {
			mockServer.expect(ExpectedCount.once(), requestTo(TEST_OSI_POSTFACH_URI))
					.andExpect(method(HttpMethod.POST))
					.andRespond(withSuccess());
		}

		private void callGrpcEndpoint() {
			grpcService.sendPostfachMail(buildSendPostfachMailRequestWithoutAttachments(), responseObserver);
		}

		private GrpcSendPostfachMailRequest buildSendPostfachMailRequestWithoutAttachments() {
			var grpcPostfachMail = GrpcPostfachMailTestFactory.createBuilder().clearId().clearAttachment().clearCreatedAt().clearDirection()
					.setVorgangId(MessageTestFactory.VORGANG_ID).build();
			return GrpcSendPostfachMailRequestTestFactory.createBuilder().setMail(grpcPostfachMail).build();
		}

		private void callResendGrpcEndpoint(String mailId) {
			grpcService.resendPostfachMail(GrpcResendPostfachMailRequestTestFactory.createBuilder().setPostfachMailId(mailId).build(),
					resendResponseObserver);
		}
	}

	@Nested
	class TestReceivingMail {

		@Autowired
		private PostfachService service;

		private void mockServerReceiveReply(String replyJson) {
			mockServer.expect(ExpectedCount.once(), requestTo(TEST_OSI_POSTFACH_URI))
					.andExpect(method(HttpMethod.GET))
					.andRespond(withSuccess(replyJson, MediaType.APPLICATION_JSON));

			mockServer.expect(ExpectedCount.once(), method(HttpMethod.DELETE)).andRespond(withSuccess());
		}

		@Nested
		class TestFetchAndListReply {

			@Autowired
			private FileService fileService;

			@Test
			void shouldListNachrichten() {
				mockServerReceiveReply(MessageJsonReplyTestFactory.buildReplyJson());

				fetchAndPersistReplies();

				var mails = callGrpcListEndpoint();
				assertThat(mails).hasSize(1);
				assertThat(mails.get(0).getCreatedAt()).isNotNull();
				assertThat(mails.get(0).getDirection()).isEqualTo(GrpcDirection.IN);

				mockServer.verify();
			}

			@Test
			void shouldContainAttachment() {
				mockServerReceiveReply(MessageJsonReplyTestFactory.buildReplyJson());

				fetchAndPersistReplies();

				var mails = callGrpcListEndpoint();
				assertThat(mails).hasSize(1);
				assertThat(mails.get(0).getAttachmentList()).hasSize(1).doesNotContainNull();
			}

			@Test
			void shouldHaveContentTypeDOCX() {
				var attachment = MessageAttachmentTestFactory.createBuilder().fileName("wort-datei.docx").build();
				mockServerReceiveReply(MessageJsonReplyTestFactory.buildReplyJson(MessageTestFactory.create(), attachment));

				fetchAndPersistReplies();

				var mails = callGrpcListEndpoint();
				assertThat(mails).hasSize(1);
				var binaryFile = fileService.findFilesMetaData(List.of(FileId.from(mails.get(0).getAttachmentList().get(0)))).findFirst()
						.orElse(OzgFile.builder().build());
				assertThat(binaryFile.getContentType()).isEqualTo("application/vnd.openxmlformats-officedocument.wordprocessingml.document");
			}

			private void fetchAndPersistReplies() {
				service.fetchAndPersistReplies();

				verify(commandService, timeout(60000)).setCommandFinished(any(), any());
			}
		}

	}

	private List<GrpcPostfachMail> callGrpcListEndpoint() {
		grpcService.findPostfachMails(GrpcFindPostfachMailsRequest.newBuilder().setVorgangId(MessageTestFactory.VORGANG_ID).build(),
				responseObserver);

		verify(responseObserver, atLeastOnce()).onNext(responseCaptor.capture());
		return responseCaptor.getValue().getMailsList();
	}

	private boolean hasAtLeastOneElement(Collection<GrpcPostfachMail> mails) {
		return !mails.isEmpty();
	}
}
