/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.StreamObserver;

class UploadStreamObserverTest {

	@Mock
	private StreamObserver<GrpcUploadBinaryFileResponse> responseObserver;
	@Mock
	private FileService service;
	@Mock
	private PolicyService policyService;

	private GrpcUploadBinaryFileRequest fileUploadRequest;
	private UploadStreamObserver uploadStreamObserver;

	@Nested
	class TestSteamObserver {

		private PipedInputStream pipedInputStream = new PipedInputStream();
		private PipedOutputStream pipedOutputStream = new PipedOutputStream();

		@BeforeEach
		void init() {
			uploadStreamObserver = new UploadStreamObserver(responseObserver, service, policyService);
			fileUploadRequest = GrpcUploadBinaryFileRequestTestFactory.createBuilder()
					.setFileContent(ByteString.copyFrom(IncomingFileTestFactory.CONTENT)).build();
		}

		@Nested
		class TestInitializing {

			@Test
			void shouldConnectPipedStreams() {
				assertDoesNotThrow(() -> uploadStreamObserver.connectInputStreamToOutputStream(pipedInputStream, pipedOutputStream));
			}

			@Test
			void shouldWriteToPipedOutputStream() {
				uploadStreamObserver.connectInputStreamToOutputStream(pipedInputStream, pipedOutputStream);

				assertDoesNotThrow(() -> pipedOutputStream.write(7574557));
			}

			@Test
			void shouldInitPipedStreams() throws Exception {
				uploadStreamObserver.initPipedStreams();

				PipedOutputStream pipedOutputStream = uploadStreamObserver.getPipedOutput();
				uploadStreamObserver.connectInputStreamToOutputStream(uploadStreamObserver.getPipedInput(), pipedOutputStream);

				assertDoesNotThrow(() -> pipedOutputStream.write(7574557));
			}

			@Test
			void shouldWriteFileContentToPipedOutputError() {
				uploadStreamObserver.initPipedStreams();

				uploadStreamObserver.storeFileContent(fileUploadRequest, uploadStreamObserver.getPipedOutput());

				assertThatThrownBy(() -> uploadStreamObserver.getPipedInput().read()).isInstanceOf(IOException.class);
			}

			@Test
			void shouldWriteFileContentToPipedOutput() throws IOException, InterruptedException, ExecutionException, TimeoutException {
				initPipedStreams();

				CompletableFuture<Integer> resFuture = writeFileContent();

				assertThat(resFuture.get(30, TimeUnit.SECONDS) > 0).isTrue();

			}

			private void initPipedStreams() {
				uploadStreamObserver.initPipedStreams();
				uploadStreamObserver.connectInputStreamToOutputStream(uploadStreamObserver.getPipedInput(), uploadStreamObserver.getPipedOutput());
			}

			private CompletableFuture<Integer> writeFileContent() {
				uploadStreamObserver.storeFileContent(fileUploadRequest, uploadStreamObserver.getPipedOutput());

				CompletableFuture<Integer> resFuture = CompletableFuture.supplyAsync(() -> {
					try {
						byte[] buffer = new byte[IncomingFileTestFactory.CONTENT.length];
						return uploadStreamObserver.getPipedInput().read(buffer);
					} catch (IOException e) {
						return -1;
					}
				});
				return resFuture;
			}

			@Test
			void shouldHandleException() {
				final IOException iOException = new IOException("test");
				assertThrows(TechnicalException.class, () -> uploadStreamObserver.handleException(iOException));

				verify(responseObserver).onNext(any());
			}
		}

		@Nested
		class TestStoreMetaData {

			@BeforeEach
			void init() {
				when(service.uploadFileStreamAsync(any(), any(), any(), any())).thenReturn(CompletableFuture.completedFuture(FileId.createNew()));
			}

			@Test
			void shouldInitPipedStreamsAndStoreMetadata() throws TimeoutException, InterruptedException, ExecutionException {
				var fileIdFuture = uploadStreamObserver.storeMetaData(fileUploadRequest);

				assertThat(fileIdFuture.get(30, TimeUnit.SECONDS)).isNotNull();
			}

			@Test
			void shouldStoreMetadata() throws InterruptedException, ExecutionException, TimeoutException {
				var fileIdFuture = uploadStreamObserver.storeMetaData(fileUploadRequest);

				assertThat(fileIdFuture.get(30, TimeUnit.SECONDS)).isNotNull();
			}

			@Test
			void shouldStoreFileMetadata() throws InterruptedException, ExecutionException, TimeoutException {
				uploadStreamObserver.initPipedStreams();
				uploadStreamObserver.storeFileMetaData(fileUploadRequest, uploadStreamObserver.getPipedInput(),
						uploadStreamObserver.getPipedOutput());

				verify(service).uploadFileStreamAsync(any(), any(), any(), any());
			}
		}

		@Nested
		class TestBuildUploadFilesReferences {

			@Test
			void shouldContainsVorgangId() {
				var uploadFilesReference = buildUploadFilesReferences();

				assertThat(uploadFilesReference.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}

			@Test
			void shouldContainsName() {
				var uploadFilesReference = buildUploadFilesReferences();

				assertThat(uploadFilesReference.getName()).isEqualTo(UploadedFilesReferenceTestFactory.NAME);
			}

			@Test
			void shouldContainsClient() {
				var uploadFilesReference = buildUploadFilesReferences();

				assertThat(uploadFilesReference.getClient()).isEqualTo(GrpcBinaryFileService.CLIENT);
			}

			private UploadedFilesReference buildUploadFilesReferences() {
				return uploadStreamObserver.buildUploadFilesReferences(GrpcUploadBinaryFileRequestTestFactory.create());
			}
		}

		@Nested
		class TestBuildOzgFile {

			@Test
			void shouldContainsName() {
				var ozgFile = buildOzgFile();

				assertThat(ozgFile.getName()).isEqualTo(OzgFileTestFactory.NAME);
			}

			@Test
			void shouldContainsSize() {
				var ozgFile = buildOzgFile();

				assertThat(ozgFile.getSize()).isEqualTo(OzgFileTestFactory.SIZE);
			}

			@Test
			void shouldContainsContentType() {
				var ozgFile = buildOzgFile();

				assertThat(ozgFile.getContentType()).isEqualTo(OzgFileTestFactory.CONTENT_TYPE);
			}

			private OzgFile buildOzgFile() {
				return uploadStreamObserver.buildOzgFile(GrpcUploadBinaryFileRequestTestFactory.create());
			}
		}
	}
}
